/**
 * Character Device Driver Sample:
 * file name: smsg.c
 */
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <asm/uaccess.h>

static int driver_major_no = 0;
static char *msg = "Hello, everyone.";
MODULE_PARM(msg, "s");

/* デバイスファイルオープン時に実行 */
static int smsg_open(struct inode *inode, struct file *filp)
{
	printk("smsg_open\n");
	return 0;
}

/* デバイスファイル読み取り時に実行 */
static int smsg_read(struct file *filp, char *buff, size_t count, loff_t *pos)
{
	int len;
	printk("smsg_read: msg = %s\n", msg);
	len = strlen(msg);
	copy_to_user(buff, msg, len);
	return 0;
}

/* デバイスファイルクローズ時に実行 */
static int smsg_release(struct inode *inode, struct file *filp)
{
	printk("smsg_release\n");
	return 0;
}

/* ファイル操作定義構造体 */
static struct file_operations driver_fops = {
	.read    = smsg_read,
	.open    = smsg_open,
	.release = smsg_release,
};

/* インストール時に実行 */
int init_module(void)
{
	int ret;
	printk("smsg: init_module: msg = %s\n", msg);

	/* キャラクタ型ドライバ管理テーブルへ登録 */
	ret = register_chrdev(driver_major_no, "smsg", &driver_fops);

	/* 登録エラー */
	if (ret < 0) {
		printk("smsg: Major no. cannot be assigned.\n");
		return ret;
	}

	/* 最初に登録する場合 */
	if (driver_major_no == 0) {
		driver_major_no = ret;
		printk("smsg: Major no. is assigned to %d.\n", ret);
	}

	return 0;
}

/* アンインストール時に実行 */
void cleanup_module(void)
{
	printk("smsg: cleanup_module\n");

	/* キャラクタ型ドライバ管理テーブルから削除 */
	unregister_chrdev(driver_major_no, "smsg");
}
